from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from typing import override

from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.model_service import (
    CreatedModelServiceEndpointMetaContext,
    CreatedModelServiceTokenContext,
)
from ai.backend.test.templates.template import (
    WrapperTestTemplate,
)

_TOKEN_DURATION = "3600"  # 1 hour


class ModelServiceTokenTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "generate_model_service_token"

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        client_session = ClientSessionContext.current()
        service_meta = CreatedModelServiceEndpointMetaContext.current()

        result = await client_session.Service(service_meta.service_id).generate_api_token(
            duration=_TOKEN_DURATION
        )
        assert "token" in result, (
            f"Token generation failed, 'token' not in result, Actual result: {result}"
        )
        token = result["token"]

        with CreatedModelServiceTokenContext.with_current(token):
            yield



from ai.backend.client.exceptions import BackendAPIError
from ai.backend.client.output.fields import image_fields
from ai.backend.common.types import ClusterMode
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.image import ImageContext
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.templates.template import TestCode
from ai.backend.test.utils.exceptions import UnexpectedSuccess


class SessionCreationFailureLowResources(TestCode):
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        image_dep = ImageContext.current()
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        session_name = f"test_failure_{str(test_id)}"

        result = await client_session.Image.get(
            image_dep.name, image_dep.architecture, fields=[image_fields["labels"]]
        )
        labels = result["labels"]
        min_mem_label = next(
            filter(lambda label: label["key"] == "ai.backend.resource.min.mem", labels), None
        )

        if not min_mem_label:
            raise RuntimeError(
                "Minimum memory label not found in image labels. Please check the image configuration."
            )

        min_mem = min_mem_label["value"]

        try:
            await client_session.ComputeSession.get_or_create(
                image_dep.name,
                architecture=image_dep.architecture,
                name=session_name,
                # This failed due to the need for additional space for shared memory.
                resources={"cpu": 1, "mem": min_mem},
                cluster_mode=ClusterMode.SINGLE_NODE,
                cluster_size=1,
            )
            raise UnexpectedSuccess("Expected BackendAPIError for low resources was not raised")
        except BackendAPIError as e:
            assert e.status == 400
            assert e.data["error_code"] == "api_generic_invalid-parameters"



from ai.backend.test.contexts.context import BaseTestContext, ContextName
from ai.backend.test.tester.dependency import ScalingGroupDep


class ScalingGroupContext(BaseTestContext[ScalingGroupDep]):
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.SCALING_GROUP



from typing import override

from ai.backend.client.exceptions import BackendAPIError
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.image import ImageContext
from ai.backend.test.contexts.session import (
    ClusterContext,
    CreatedSessionMetaContext,
    SessionContext,
)
from ai.backend.test.templates.template import TestCode


class SessionRenameFailureDuplicatedName(TestCode):
    @override
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        session_meta = CreatedSessionMetaContext.current()
        session_name = session_meta.name

        image_dep = ImageContext.current()
        cluster_dep = ClusterContext.current()
        session_dep = SessionContext.current()
        second_session_name = session_name + "_second"

        # TODO: Make second session creation/deletion with template
        await client_session.ComputeSession.get_or_create(
            image_dep.name,
            resources=session_dep.resources,
            type_="interactive",
            name=second_session_name,
            cluster_mode=cluster_dep.cluster_mode,
            cluster_size=cluster_dep.cluster_size,
        )
        try:
            await client_session.ComputeSession(name=second_session_name).rename(session_name)
            raise AssertionError(
                "Renaming a session to a name that already exists should raise an error"
            )
        except BackendAPIError:
            pass
        finally:
            await client_session.ComputeSession(name=second_session_name).destroy()



from ai.backend.client.session import AsyncSession


async def retrieve_all_files(
    client_session: AsyncSession, vfolder_name: str, path: str = ""
) -> set[str]:
    """
    Recursively retrieves all file paths in a virtual folder.
    """

    response = await client_session.VFolder(vfolder_name).list_files(path)
    assert "items" in response, "Response does not contain 'items' key."

    all_files = set()

    for item in response["items"]:
        if item["type"] == "FILE":
            file_path = f"{path}/{item['name']}" if path else item["name"]
            all_files.add(file_path)
        elif item["type"] == "DIRECTORY":
            subdir_path = f"{path}/{item['name']}" if path else item["name"]
            subdir_files = await retrieve_all_files(client_session, vfolder_name, subdir_path)
            all_files.update(subdir_files)

    return all_files



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from typing import override

from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.contexts.vfolder import CreatedVFolderMetaContext, VFolderContext
from ai.backend.test.data.vfolder import VFolderMeta
from ai.backend.test.templates.template import WrapperTestTemplate


class UserVFolderTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "general_user_vfolder"

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        client_session = ClientSessionContext.current()
        vfolder_name = f"test-{str(test_id)[:8]}"
        vfolder_cfg = VFolderContext.current()

        vfolder = None
        try:
            vfolder = await client_session.VFolder.create(
                name=vfolder_name,
                unmanaged_path=vfolder_cfg.unmanaged_path,
                permission=vfolder_cfg.permission,
                usage_mode="general",
                cloneable=vfolder_cfg.cloneable,
            )

            with CreatedVFolderMetaContext.with_current(
                VFolderMeta(id=vfolder["id"], name=vfolder_name)
            ):
                yield
        finally:
            if vfolder:
                await client_session.VFolder.delete_by_id(vfolder["id"])


class ProjectVFolderTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "general_project_vfolder"

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        client_session = ClientSessionContext.current()
        vfolder_name = f"test-{str(test_id)[:8]}"
        vfolder_cfg = VFolderContext.current()

        vfolder = None
        try:
            vfolder = await client_session.VFolder.create(
                name=vfolder_name,
                group=vfolder_cfg.group,
                unmanaged_path=vfolder_cfg.unmanaged_path,
                permission=vfolder_cfg.permission,
                usage_mode="general",
                cloneable=vfolder_cfg.cloneable,
            )

            with CreatedVFolderMetaContext.with_current(
                VFolderMeta(id=vfolder["id"], name=vfolder_name)
            ):
                yield
        finally:
            if vfolder:
                await client_session.VFolder.delete_by_id(vfolder["id"])
                await client_session.VFolder(vfolder["name"]).purge()



from pathlib import Path

from ai.backend.common.json import load_json
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.session import CreatedSessionMetaContext
from ai.backend.test.templates.template import TestCode


class FileExistenceCheck(TestCode):
    _path: Path
    _checklist: list[str]

    def __init__(self, path: Path, checklist: list[str]) -> None:
        super().__init__()
        self._path = path
        self._checklist = checklist

    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        session_meta = CreatedSessionMetaContext.current()

        response = await client_session.ComputeSession(str(session_meta.id)).list_files(self._path)
        files = load_json(response["files"])

        actual_files = [file["filename"] for file in files if file["filename"] in self._checklist]
        assert actual_files == self._checklist, (
            f"Expected files: {self._checklist}, but got following files: {files}"
        )



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from typing import override

from ai.backend.test.contexts.client_session import (
    ClientSessionContext,
    CreatedUserClientSessionContext,
)
from ai.backend.test.contexts.user import CreatedUserContext
from ai.backend.test.contexts.vfolder import (
    CreatedVFolderMetaContext,
    VFolderInvitationContext,
    VFolderInvitationPermissionContext,
)
from ai.backend.test.data.vfolder import VFolderInvitationMeta
from ai.backend.test.templates.template import WrapperTestTemplate


class VFolderInviteTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "vfolder_invite"

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()
        test_user = CreatedUserContext.current()
        perm = VFolderInvitationPermissionContext.current()

        response = await client_session.VFolder(vfolder_meta.name).invite(
            perm=perm, emails=[test_user.email]
        )
        assert "invited_ids" in response, (
            f"Response does not contain 'invited_ids', Actual value: {response}"
        )
        invited_ids = response["invited_ids"]

        invitation_meta = VFolderInvitationMeta(
            invited_user_emails=invited_ids,
        )

        with VFolderInvitationContext.with_current(invitation_meta):
            yield


class AcceptInvitationTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "accept_invitation"

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        client_session = CreatedUserClientSessionContext.current()
        created_user_meta = CreatedUserContext.current()
        mount_permission = VFolderInvitationPermissionContext.current()
        invitation_response = await client_session.VFolder.invitations()
        assert len(invitation_response["invitations"]) > 0, "No invitations found"
        invitation = invitation_response["invitations"][0]
        assert "id" in invitation, "Invitation ID not found in the response"
        assert invitation["status"] == "pending", (
            f"Invitation is not pending. Actual value: {invitation['status']}"
        )
        assert invitation["invitee_user_email"] == created_user_meta.email, (
            f"Invitation email does not match created user email. Actual value: {invitation['inviter_user_email']}"
        )
        assert invitation["mount_permission"] == mount_permission, (
            f"Invitation mount permission does not match. Actual value: {invitation['mount_permission']}"
        )

        await client_session.VFolder.accept_invitation(invitation["id"])
        yield


class RejectInvitationTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "reject_invitation"

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        client_session = CreatedUserClientSessionContext.current()
        created_user_meta = CreatedUserContext.current()
        mount_permission = VFolderInvitationPermissionContext.current()
        invitation_response = await client_session.VFolder.invitations()
        assert len(invitation_response["invitations"]) > 0, "No invitations found"
        invitation = invitation_response["invitations"][0]
        assert "id" in invitation, "Invitation ID not found in the response"
        assert invitation["status"] == "pending", (
            f"Invitation is not pending. Actual value: {invitation['status']}"
        )
        assert invitation["invitee_user_email"] == created_user_meta.email, (
            f"Invitation email does not match created user email. Actual value: {invitation['inviter_user_email']}"
        )
        assert invitation["mount_permission"] == mount_permission, (
            f"Invitation mount permission does not match. Actual value: {invitation['mount_permission']}"
        )

        await client_session.VFolder.delete_invitation(invitation["id"])
        yield



from ai.backend.client.exceptions import BackendAPIError
from ai.backend.test.contexts.client_session import (
    ClientSessionContext,
    CreatedUserClientSessionContext,
)
from ai.backend.test.contexts.user import CreatedUserContext
from ai.backend.test.contexts.vfolder import (
    CreatedVFolderMetaContext,
    VFolderInvitationPermissionContext,
)
from ai.backend.test.templates.template import TestCode
from ai.backend.test.utils.exceptions import UnexpectedSuccess


class VFolderInviteFailure(TestCode):
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()
        test_user = CreatedUserContext.current()
        perm = VFolderInvitationPermissionContext.current()

        try:
            await client_session.VFolder(vfolder_meta.name).invite(
                perm=perm, emails=[test_user.email]
            )
        except BackendAPIError as e:
            assert e.status == 404, f"Expected 404 error, but got {e.status}: Exception: {e}"


class VFolderAcceptDuplicatedInvitation(TestCode):
    async def test(self) -> None:
        client_session = CreatedUserClientSessionContext.current()
        created_user_meta = CreatedUserContext.current()
        mount_permission = VFolderInvitationPermissionContext.current()
        invitation_response = await client_session.VFolder.invitations()
        assert len(invitation_response["invitations"]) > 0, "No invitations found"
        invitation = invitation_response["invitations"][0]
        assert "id" in invitation, "Invitation ID not found in the response"
        assert invitation["status"] == "pending", (
            f"Invitation is not pending. Actual value: {invitation['status']}"
        )
        assert invitation["invitee_user_email"] == created_user_meta.email, (
            f"Invitation email does not match created user email. Actual value: {invitation['inviter_user_email']}"
        )
        assert invitation["mount_permission"] == mount_permission, (
            f"Invitation mount permission does not match. Actual value: {invitation['mount_permission']}"
        )

        await client_session.VFolder.accept_invitation(invitation["id"])

        try:
            await client_session.VFolder.accept_invitation(invitation["id"])
            raise UnexpectedSuccess("Invitation was accepted again, but it should have failed.")
        except BackendAPIError as e:
            assert e.status == 404, f"Expected 404 error, but got {e.status}: Exception: {e}"



import json
from contextlib import closing

from ...utils.cli import EOF, ClientRunnerFunc, decode


def test_alias_image(run_admin: ClientRunnerFunc):
    pass


def test_dealias_image(run_admin: ClientRunnerFunc):
    pass


def test_list_image(run_admin: ClientRunnerFunc):
    print("[ List image ]")
    with closing(run_admin(["--output=json", "admin", "image", "list"])) as p:
        p.expect(EOF)
        decoded = decode(p.before)
        loaded = json.loads(decoded)
        image_list = loaded.get("items")
        assert isinstance(image_list, list), "Image list not printed properly"



from ai.backend.client.session import AsyncSession
from ai.backend.test.contexts.context import BaseTestContext, ContextName


class ClientSessionContext(BaseTestContext[AsyncSession]):
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.CLIENT_SESSION


class CreatedUserClientSessionContext(BaseTestContext[AsyncSession]):
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.CREATED_USER_CLIENT_SESSION
